/*
 * bloom.h:
 * 
 * bloom filter for NIST NSRL
 * Originally (C) August 2006, Simson L. Garfinkel
 * Released into the public domain in September 2008.
 */

#ifndef NSRL_BLOOM_H
#define NSRL_BLOOM_H

#ifndef PACKAGE
#error must include autoconf config.h file
#endif

#include <stdio.h>
#include <string.h>
#include <sys/types.h>

#ifdef _MSC_VER
typedef unsigned int u_int;
typedef u_char u_char;
typedef unsigned __int64 u_int64_t;
typedef unsigned __int64 uint64_t;
typedef __int64 int64_t;
#endif

#ifdef HAVE_OPENSSL_HMAC_H
#include <openssl/hmac.h>		// includes EVP
#endif

#ifdef __cplusplus
extern "C" {
#endif
#ifdef NEVER_DEFINED
}
#endif

typedef struct  nsrl_bloom_ {
    u_int hash_bits;            // how many bits in the input hash (128 or 160)
    u_int hash_bytes;		// hash_bits/8 ; performance optimization
    u_int M;	                // number of bits of hash to use for each bloom function (log2(vector_bytes))
    u_int k;	                // number of bloom functions to use
    size_t vector_bytes;		// length of the vector (vector_bytes*8 = 2^M)
    u_int vector_offset;	// offset of vector in file (usually 4096)
    u_char *vector;		// the bloom filter
    char *comment;
    u_int64_t added_items;		// How many items have been added.
    int   fd;			// file descriptor for mapped or open file
    u_int debug;		// debug level
    u_int memmapped:1;		// true if we need to unmap; otherwise we free(vector)
    u_int free_this:1;		// should we free this on clean?
    u_int fileio:1;				// force file i/o
    u_int64_t hits;			// stats

#ifdef HAVE_OPENSSL_HMAC_H
    /* Encryption Support; requires OpenSSL */
    u_char *key;			// hash of passphrase; 0 for no encryption
    const EVP_MD *md;
#endif
} nsrl_bloom;



/* Service functions for hexdecimal conversion: */
int nsrl_hex2bin(u_char *binbuf,size_t binbuf_size, const char *hex);
const char *nsrl_hexbuf(char *dst,int dst_len,const u_char *bin,int bytes,int flag);

/** nsrl_bloom_alloc(): allocates the filter memory. */
nsrl_bloom *nsrl_bloom_alloc();		// allocate a new bloom object

/** nsrl_bloom_create:
 * Create a nsrl bloom filter for either a 128-bit or a 160-bit hash.
 * If fname is specified, the file is created; otherwise the bloom
 * filter is kept in memory. M is log2(m); k is the number of functions to use.
 * Returns 0 if successful, -1 if failure.
 */

int nsrl_bloom_create(nsrl_bloom *b,const char *fname,int hash_size, int M, int k, const char *comment);

/** nsrl_bloom_open: 
 * Opens an existing bloom filter, taking the parameters from it.
 * Returns 0 if successful, -1 if failure.
 */

int nsrl_bloom_open(nsrl_bloom *b,const char *fname,int mode);	// open an existing filter
int nsrl_bloom_write(nsrl_bloom *b,const char *fname);	// open an existing filter

/* add and query */

void nsrl_bloom_fprint_info(const nsrl_bloom *b,FILE *out);
void nsrl_bloom_print_info(const nsrl_bloom *b);
void nsrl_bloom_fprint_usage(FILE *out);
void nsrl_bloom_print_usage();
void nsrl_bloom_add(nsrl_bloom *b,const u_char *hash); 
void nsrl_bloom_addString(nsrl_bloom *b,const char *str); 
int  nsrl_bloom_query( nsrl_bloom *b,const u_char *hash); // 0 not present; 1 is present
int  nsrl_bloom_queryString( nsrl_bloom *b,const char *str); // 0 not present; 1 is present
double nsrl_bloom_utilization(const nsrl_bloom *b);			  // on scale 0..1

/* Sets an encryption key */
void nsrl_set_passphrase(nsrl_bloom *b,const char *passphrase);

void  nsrl_bloom_clear(nsrl_bloom *b);	// Release resources and sanitize
void  nsrl_bloom_free(nsrl_bloom *b);	// free *b


/* debug */
void nsrl_bloom_info(char *buf,int buflen,const nsrl_bloom *b);	
void nsrl_calc_histogram(const nsrl_bloom *b,u_int counts[256]);
void nsrl_print_histogram(const nsrl_bloom *b,const u_int counts[256]);


#define NSRL_HEXBUF_UPPERCASE 0x01
#define NSRL_HEXBUF_SPACE2    0x02
#define NSRL_HEXBUF_SPACE4    0x04

#ifdef NEVER_DEFINED
{
#endif
#ifdef __cplusplus
}
#endif

#ifdef __cplusplus
#include <string>
class NSRLBloom : public nsrl_bloom {
public:
    NSRLBloom(){
	this->vector = 0;
	this->fd = 0;
	this->comment = 0;
#ifdef HAVE_OPENSSL_HMAC_H
	this->key = 0;
#endif
	nsrl_bloom_clear(this);
    };
    int open(const char *fname,int mode){
	return nsrl_bloom_open(this,fname,mode);
    }
    int create(const char *fname,int hash_size, int M, int k, const char *comment){
	return nsrl_bloom_create(this,fname,hash_size,M,k,comment);
    }
    void print_info() const {
	nsrl_bloom_print_info(this);
    }

    void fprint_info(FILE *out) const {
	nsrl_bloom_fprint_info(this,out);
    }

    void print_usage() const {
	nsrl_bloom_print_usage();
    }

    void add(const u_char *hash){
	nsrl_bloom_add(this,hash);
    }
    void addString(const char *str){
	nsrl_bloom_addString(this,str);
    }
    void addString(const std::string &str){
	nsrl_bloom_addString(this,str.c_str());
    }
    void addHex(const std::string &str){
	u_char buf[256];		
	nsrl_hex2bin(buf,sizeof(buf),str.c_str());
	add(buf);
    }
    bool query(const u_char *hash) {// true is present; false is not present
	return nsrl_bloom_query(this,hash);
    }
#ifdef HAVE_OPENSSL_HMAC_H
    bool queryString(const char *str)  {// true is present; false is not present
	return nsrl_bloom_queryString(this,str);
    }
    bool queryString(const std::string &str)  {// true is present; false is not present
	return nsrl_bloom_queryString(this,str.c_str());
    }
#endif
    double utilization() const{
	return nsrl_bloom_utilization(this);
    }
    void set_passphrase(const char *passphrase){
	nsrl_set_passphrase(this,passphrase);
    }
    unsigned long long calchits(){
	return this->hits;
    }
    int write(const char *fname){
	return nsrl_bloom_write(this,fname);
    }

    ~NSRLBloom(){
	nsrl_bloom_clear(this);
    }
};
#endif
#endif
