/**
 * fiwalk.cpp:
 * File and Inode Walk.
 *
 * This application uses SleuthKit to generate a report of all of the files
 * and orphaned inodes found in a disk image. It can optionally compute the
 * MD5 of any objects, save those objects into a directory, or both.
 *
 * Algorithm:
 * 1 - Find all of the partitions on the disk. 
 * 2 - For each partition, walk the files.
 * 3 - For each file, print the requested information.
 * 4 - For each partition, walk the indoes
 * 5 - For each inode, print the requested information.
 *
 * @author Simson Garfinkel
 */


/* config.h must be first */
#include "config.h"

#include <strings.h>
#include <string.h>
#include <stdlib.h>
#include <stdarg.h>

#include <sys/types.h>
#include <sys/time.h>
#include <sys/stat.h>
#include <pwd.h>

#include <regex.h>
#include <ctype.h>
#include <fcntl.h>

#ifdef HAVE_STDINT_H
#include <stdint.h>
#endif

#ifdef HAVE_SIGNAL_H
#include <sys/signal.h>
#endif

#ifdef HAVE_SYS_RESOURCE_H
#include <sys/resource.h>
#endif

#ifdef HAVE_SYS_UTSNAME_H
#include <sys/utsname.h>
#endif

#ifdef HAVE_INTTYPES_H
#ifndef __STDC_FORMAT_MACROS
#define __STDC_FORMAT_MACROS
#endif
#include <inttypes.h>
#endif

#ifdef HAVE_ERR_H
#include <err.h>
#endif

#include <tsk3/libtsk.h>

#ifdef HAVE_LIBAFFLIB
#include <afflib/afflib.h>
#include <afflib/afflib_i.h>
#include <afflib/utils.h>		// aff::seglist
#endif

/* Bring in our headers */
#include "fiwalk.h"
#include "content.h"
#include "arff.h"
#include "plugin.h"
#include "bloom.h"
#include "bloomset.h"

#include "threaded_hash.h"
#include "xml.h"
#include "base64.h"


/* Output Devices */
class arff *a = 0;			// ARFF generator
class xml  *x = 0;
FILE  *t = 0;				// text output or body file enabled

/* Configurable options */
bool opt_multithreaded = false;		// use multiple threading

/* Runtime options */
string save_outdir = ".";
int  opt_debug    = 0;
bool opt_mactime  = false;
bool opt_md5      = true;		// calculate the MD5 for every file?
bool opt_sha1     = true;		// calculate the SHA1 for every file?
bool opt_save     = false;
bool opt_get_fragments = false;		// get the fragments
bool opt_no_orphans = false;
bool opt_body_file = false;
const char *config_file = 0;
int  file_count_max = 0;
int  file_count = 0;
int  next_id = 1;

u_int sector_size=512;			// defaults to 512
int  opt_M = 30;
int  opt_k = 4;


/* Bloomset is used for suppression */
bloomset mybloomset;

/* Sector hash bloom calculations */
u_int  sectorhash_size=512;
bool  opt_compute_sector_hashes = false;
bool  opt_print_sector_hashes = false;
NSRLBloom *sector_bloom = 0;

namelist_t namelist;		// names of files that we want to find

#ifdef HAVE_LIBMAGIC
bool opt_magic = true;		// get the output of libmagic by default
#else
bool opt_magic = false;		// don't get the output of the file command
#endif


/****************************************************************
 ** Current information.
 ****************************************************************/
int      current_partition_num=0;
int64_t  current_partition_start=0;	// in bytes

/* Individual 'state' variables */
string  plugin_filename;

void print_version()
{
    printf("FIWalk Version:    %s\n",PACKAGE_VERSION);
    printf("SleuthKit Version: %s\n",tsk_version_get_str());
#ifdef HAVE_LIBAFFLIB
    printf("AFFLIB Version:    %s\n",af_version());
#else
    printf("*** NO AFFLIB SUPPORT ***\n");
#endif    
    
}

#ifndef HAVE_ERR
#include <stdarg.h>
void err(int eval,const char *fmt,...)
{
    va_list ap;
    va_start(ap,fmt);
    vfprintf(stderr,fmt,ap);
    va_end(ap);
    fprintf(stderr,": %s\n",strerror(errno));
    exit(eval);
}
#endif

void usage()
{
    printf("usage: fiwalk [options] iso-name\n");
    printf("Default behavior: Just print the file system statistics and exit.\n");
    printf("options:\n");
    printf("    -c config.txt   read config.txt for metadata extraction tools\n");
    printf("    -C nn           only process nn files, then do a clean exit\n"); 
    printf("    -u              Guess at usernames based on filenames, mode, or UIDs\n");
    
    printf("\n");
    printf("include/exclude parameters; may be repeated. \n");
    printf("    -b+bloomfile.bloom = Only include files whose SHA1 is in bloomfile.bloom\n");
    printf("    -b-bloomfile.bloom = Only include files whose SHA1 is not in bloomfile.bloom\n");
    printf("    -n pattern  = only match files for which the filename matches\n");
    printf("                  the pattern.\n");
    printf("              example: -n .jpeg -n .jpg will find all JPEG files\n");
    printf("              Case is ignored. Will not match orphan files.\n");
    printf("    ");
    printf("\n");
    printf("Ways to make this program run faster:\n");
    printf("    -g just report the file fragments - don't get the data\n");
    printf("    -O do not walk the unallocated or orphan files\n");
    printf("    -z do not calculate MD5 or SHA1 values\n");
    printf("    -2 multi-threaded hash calculation (experimental).\n");
#ifdef HAVE_LIBMAGIC
    printf("    -F = Disable LIBMAGIC (enabled by default)");
#else
    printf("    -f = Report the output of the 'file' command for each\n");
#endif
    printf("\n");
    printf("Full content options:\n");
    printf("    -m = Output in SleuthKit 'Body file' format\n");
    printf("    -M = Report MD5 for each file (default on)\n");
    printf("    -1 = Report SHA1 for each file (default on)\n");
    printf("    -s <dir> = Save all recovered files to <dir>\n");
    printf("    -a <audit.txt> = Read the scalpel audit.txt file\n");
    printf("\n");
    printf("Output options:\n");
    printf("    -A<file> = ARFF output to <file>\n");
    printf("    -X<file> = XML output to a <file> (full DTD)\n");
    printf("    -x       = XML output to stdout (no DTD)\n");
    printf("    -T<file> = Walkfile output to <file>\n");
    printf("\n");
    printf("Sector hash:\n");
    printf("    -E             = Print sector hashes\n");
    printf("    -Snnn          = Specify sector hash size; default is %d\n",sectorhash_size);
    printf("    -B<file>       = output sector hash bloom filter to <file>; M=%d k=%d\n",opt_M,opt_k);
    
    printf("Misc:\n");
    printf("    -d = debug this program\n");
    printf("    -v = Enable SleuthKit verbose flag\n");
    printf("\n");
    print_version();
    exit(1);
}

/****************************************************************
 ** Support routines 
 ****************************************************************/
static const char *cstr(const string &str){
    return str.c_str();
}

static string empty("");


/****************************************************************
 ** XML output
 ****************************************************************/

/****************************************************************
 ** Metadata Output
 ****************************************************************/

/**
 * output a comment in the current file format
 */
void comment(const char *format,...)
{
    char buf[1024];
    va_list ap;
    va_start(ap, format);
    vsnprintf(buf,sizeof(buf),format,ap);
    va_end(ap);

    if(t) fprintf(t,"# %s\n",buf);
    if(x) x->comment(buf);
    if(a) a->add_comment(buf);
}

/**
 * output a name/format/value for the current partition.
 * This information is simply printed as comments for ARFF files.
 *
 * @param name - the name of the thing being output
 * @param format - the format for the thing
 * @param ... - the value
 * This will output as a comment in the ARFF file
 */
void partition_info(const string &name,const string &value,const string &attribute)
{
    if(name.find(" ")<0) err(1,"partition_info(%s) has a space in it",cstr(name));
    if(a) a->add_comment(name + ": " + value);
    if(t && !opt_body_file) fputs(cstr(name + ": " + value + "\n"),t);
    if(x) x->xmlout(name,value,attribute,true);
}

void partition_info(const string &name,const string &value)
{
    partition_info(name,value,empty);
}

void partition_info(const string &name,long i)
{
    char buf[1024];
    sprintf(buf,"%ld",i);
    partition_info(name,buf,empty);
}


/****************************************************************
 * These file_info(name,value) are called for each extracted attribute
 * for each file.  Some of the class are made by metadata extraction
 * in this module; others are called from the plugin system.
 */
void file_info_xml(const string &name,const string &value)
{
    if(x){
	x->push(name);
	x->puts(value);
	x->pop();
    }
}

/* Process a string value */
void file_info(const string &name,const string &value)
{
    if(a) a->add_value(name,value); 
    if(t && !opt_body_file) fputs(cstr(name + ": " + value + "\n"),t); 
    if(x) x->xmlout(name,value); 
}

/* this file_info is for sending through a hash. */
void file_info(class threaded_hash &h)
{
    if(a) a->add_value(h.name(),h.hexhash());
    if(t && !opt_body_file) fputs(cstr(h.name() + ": " + h.hexhash() + "\n"),t);
    if(x) x->xmlout("hashdigest",h.hexhash(),"type='"+h.name()+"'",false);
}

/* Process a numeric value */
void file_info(const string name, int64_t value)
{
    if(a) a->add_value(name,value); 
    if(t || x){
	if(t) fprintf(t,"%s: %"PRId64"\n",cstr(name),value);
	if(x) x->xmlprintf(name,"","%"PRId64,value);
    }
}

/* Process a temporal value */
void file_infot(const string name,time_t t0)
{
    if(a) a->add_valuet(name,t0); 
    if(x){
	char buf[16];
	snprintf(buf,sizeof(buf),"%ld",(long)t0);
	x->xmlout(name,buf);
    }
    if(t) {
	char buf[64];
	fprintf(t,"%s: %ld\n",name.c_str(),(long)t0);
	strftime(buf,sizeof(buf),"%F %T",gmtime(&t0));
	fprintf(t,"%s_txt: %s\n",name.c_str(),buf);
    }
}

/****************************************************************/
string mytime()
{
    time_t t = time(0);
    char *buf = ctime(&t);
    buf[24] = 0;
    return string(buf);
}

/****************************************************************
 ** Support Code.
 ****************************************************************/

/* FS_INFO defined in ~/sleuthkit-2.05/src/fstools/fs_tools.h
 * IMG_INFO defined in img_tools.h 
 * img_open - opens the image; calls do_dimage
 * do_dimage - analyizes the image
 *        mm_open - opens the volume; mm_part_walk - walk each volume;
 *                  calls mm_act for each
 * mm_act - analyizes each partition; calls do_vol for ones that we can analyze
 * do_vol - analyizes each volume
 *       fs_open - opens the file system; dent_walk 
 *               - walks the directory structure.
 *               - calls dent_act for each file
 * dent_act - prints file name; calls file_walk() to read each file; 
 *          - calls file_act for each sector
 * file_act - prints the block number
 */

/** 
 * file_act:
 *  -- obtain the address of each data unit
 *  -- optionally calculate the mac or save the data
 * 
 */
TSK_WALK_RET_ENUM
file_act(TSK_FS_FILE * fs_file, TSK_OFF_T a_off, TSK_DADDR_T addr, char *buf,
	 size_t size, TSK_FS_BLOCK_FLAG_ENUM flags, void *ptr)
{
    content *ci = (content *)ptr;

    if(opt_debug>1){
	printf("file_act(fs_file=%p,addr=%"PRIuDADDR" buf=%p size=%d)\n",
	       fs_file,addr,buf,(int)size);
	if(opt_debug>1 && ci->segs.size()==0){
	    if(fwrite(buf,size,1,stdout)!=1) err(1,"fwrite");
	    printf("\n");
	}
    }

    if(size==0)  return TSK_WALK_CONT;	// can't do much with this...

    if(!opt_get_fragments){
	ci->add_bytes(buf,a_off,size);	// add these bytes to the file
    }

    /* "Address 0 is reserved in ExtX and FFS to denote a "sparse"
       block (one which is all zeros).  TSK knows this and returns
       zeros when a file refers to block 0.  You can check the 'flags'
       argument to the callback to determine if the data is from
       sparse or compressed data. RAW means that the data in the
       buffer was read from the disk.

       TSK_FS_BLOCK_FLAG_RAW
       TSK_FS_BLOCK_FLAG_SPARSE
       TSK_FS_BLOCK_FLAG_COMP"
    */

    int64_t  fs_offset = ((int64_t)addr)*fs_file->fs_info->block_size;
    int64_t img_offset = current_partition_start + fs_offset;

    if(ci->segs.size()>0){
	if((ci->segs.back().flags & TSK_FS_BLOCK_FLAG_SPARSE) 
	   && (flags & TSK_FS_BLOCK_FLAG_SPARSE)){
	    ci->segs.back().len += size;
	    return TSK_WALK_CONT;
	}

	/* See if we can extend the last segment in the segment list,
	 * or if this is the start of a new fragment.
	 */
	if((int64_t)(ci->segs.back().img_offset + ci->segs.back().len) == (int64_t)img_offset){
	    ci->segs.back().len += size;
	    return TSK_WALK_CONT;
	}
    }
    /* Need to add a new element to the list */
    ci->add_seg(img_offset,fs_offset,(int64_t)a_off,size,flags);
    return TSK_WALK_CONT;
}


/**
 * Process the contents of a file.
 *
 * @return 1 on error and 0 on success 
 */
void tsk_fs_make_ls(TSK_FS_META * a_fs_meta, char *ls)
{
    /* put the default values in */
    strcpy(ls, "----------");

    if (a_fs_meta->type < TSK_FS_META_TYPE_STR_MAX)
        ls[0] = tsk_fs_meta_type_str[a_fs_meta->type][0];

    /* user perms */
    if (a_fs_meta->mode & TSK_FS_META_MODE_IRUSR)
        ls[1] = 'r';
    if (a_fs_meta->mode & TSK_FS_META_MODE_IWUSR)
        ls[2] = 'w';
    /* set uid */
    if (a_fs_meta->mode & TSK_FS_META_MODE_ISUID) {
        if (a_fs_meta->mode & TSK_FS_META_MODE_IXUSR)
            ls[3] = 's';
        else
            ls[3] = 'S';
    }
    else if (a_fs_meta->mode & TSK_FS_META_MODE_IXUSR)
        ls[3] = 'x';

    /* group perms */
    if (a_fs_meta->mode & TSK_FS_META_MODE_IRGRP)
        ls[4] = 'r';
    if (a_fs_meta->mode & TSK_FS_META_MODE_IWGRP)
        ls[5] = 'w';
    /* set gid */
    if (a_fs_meta->mode & TSK_FS_META_MODE_ISGID) {
        if (a_fs_meta->mode & TSK_FS_META_MODE_IXGRP)
            ls[6] = 's';
        else
            ls[6] = 'S';
    }
    else if (a_fs_meta->mode & TSK_FS_META_MODE_IXGRP)
        ls[6] = 'x';

    /* other perms */
    if (a_fs_meta->mode & TSK_FS_META_MODE_IROTH)
        ls[7] = 'r';
    if (a_fs_meta->mode & TSK_FS_META_MODE_IWOTH)
        ls[8] = 'w';

    /* sticky bit */
    if (a_fs_meta->mode & TSK_FS_META_MODE_ISVTX) {
        if (a_fs_meta->mode & TSK_FS_META_MODE_IXOTH)
            ls[9] = 't';
        else
            ls[9] = 'T';
    }
    else if (a_fs_meta->mode & TSK_FS_META_MODE_IXOTH)
        ls[9] = 'x';
}


static uint8_t
process_tsk_file(TSK_FS_FILE * fs_file, const char *path)
{
    if ((fs_file->meta == NULL) || (fs_file->name == NULL)) return 1;
    if (fs_file->meta->type != TSK_FS_META_TYPE_REG) return 0;

    if (opt_debug) printf("Processing %s%s\n", path, fs_file->name->name);

    /* Recover the filename from the fs_dent, if it is provided */
    content ci;				// where the content will go
    ci.evidence_dirname = path;
    ci.set_filename(fs_file->name->name);

    /* If we are filtering and we have a filename, see if we want this file. */
    if (ci.name_filtered()) return 0;

    /* Looks like we are processing */
    if(a) a->new_row();			// tell ARFF we are starting a new row
    if(x) x->push("fileobject"); 	// tell XML we are starting a new XML object
	    
    /* Get the content if needed */
    if(ci.need_file_walk()){
	int myflags = TSK_FS_FILE_WALK_FLAG_NOID;
	if (opt_get_fragments) myflags |= TSK_FS_FILE_WALK_FLAG_AONLY;
	if (tsk_fs_file_walk (fs_file, (TSK_FS_FILE_WALK_FLAG_ENUM) myflags, file_act, (void *) &ci)) {

	    // ignore errors from deleted files that were being recovered
	    if (tsk_errno != TSK_ERR_FS_RECOVER) {
		if(opt_debug){
		    fprintf(stderr,"Processing: %s/%s (%" PRIuINUM ")\n", path,
			   fs_file->name->name, fs_file->meta->addr);
		    tsk_error_print(stderr);
		}
	    }
	    tsk_error_reset();
	}
    }

    /* See if the bloomset says to get rid of the file */
    if(mybloomset.size()>0){
	if(mybloomset.check_exclude(ci.h_sha1.hash())){
	    file_info("bloom","filtered");
	    if(x) x->pop();
	    if(t) fputs("\n",t);
	    return TSK_WALK_CONT;
	}
    }

    if(file_count_max && file_count>file_count_max) return TSK_WALK_STOP;
    file_count++;

    /* Send through to the plugin if we were doing that.
     * Currently results only go to ARFF file, not to the XML file.
     */

    /* Finally output the informaton */
    if(opt_body_file){
	char ls[64];
	tsk_fs_make_ls(fs_file->meta,ls);
	fprintf(t,"%s|%s|%"PRId64"|%s|%d|%d|%"PRId64"|%d|%d|%d|%d\n",
		ci.h_md5.hexhash(),ci.filename().c_str(),fs_file->meta->addr,
		ls,fs_file->meta->uid,fs_file->meta->gid,
		fs_file->meta->size,
		(uint32_t)(fs_file->meta->atime),
		(uint32_t)fs_file->meta->mtime,
		(uint32_t)fs_file->meta->ctime,
		(uint32_t)fs_file->meta->crtime);
	return TSK_WALK_CONT;
    }

    /* Available information in fs_file:
     * int fs_file->tag (internal)
     * TSK_FS_NAME *fs_file->name (file name)
     * TSK_FS_META *fs_file->meta
     * TSK_FS_INFO *fs_file->fs_info
     */

    if(ci.has_filename()) file_info(string("filename"),ci.filename());
    file_info("id",next_id++);
    file_info("filesize",fs_file->meta->size);
    file_info("partition",current_partition_num);
    if(fs_file->meta->flags & TSK_FS_META_FLAG_ALLOC)   file_info("alloc",1);
    if(fs_file->meta->flags & TSK_FS_META_FLAG_UNALLOC) file_info("unalloc",1);
    if(fs_file->meta->flags & TSK_FS_META_FLAG_USED)    file_info("used",1);
    if(fs_file->meta->flags & TSK_FS_META_FLAG_UNUSED)  file_info("unused",1);
    if(fs_file->meta->flags & TSK_FS_META_FLAG_ORPHAN)  file_info("orphan",1);
    if(fs_file->meta->flags & TSK_FS_META_FLAG_COMP)    file_info("compressed",1);
    file_info("inode",fs_file->meta->addr);
    file_info("type",fs_file->meta->type);
    file_info("mode",fs_file->meta->mode); // *** REPLACE WITH drwx-rw-rw or whatever 
    file_info("nlink",fs_file->meta->nlink);
    file_info("uid",fs_file->meta->uid);
    file_info("gid",fs_file->meta->gid);
    if(fs_file->meta->mtime) file_infot("mtime",fs_file->meta->mtime);
    if(fs_file->meta->ctime) file_infot("ctime",fs_file->meta->ctime);
    if(fs_file->meta->atime) file_infot("atime",fs_file->meta->atime);
    if(fs_file->meta->crtime) file_infot("crtime",fs_file->meta->crtime);
    if(TSK_FS_TYPE_ISEXT(fs_file->fs_info->ftype)){
	if(fs_file->meta->time2.ext2.dtime){
	    file_infot("dtime",fs_file->meta->time2.ext2.dtime);
	}
    }

    if(TSK_FS_TYPE_ISHFS(fs_file->fs_info->ftype)){
	if(fs_file->meta->time2.hfs.bkup_time){
	    file_infot("bkup_time",fs_file->meta->time2.hfs.bkup_time);
	}
    }

    /* TK: do content_ptr */
    if(fs_file->meta->seq!=0) file_info("seq",fs_file->meta->seq);
    /* TK: do attr */
    /* TK: do attr_state */
    /* TK: do name2 */
    if(fs_file->meta->link && fs_file->meta->link[0]!=0){
	file_info("link_target",fs_file->meta->link);
    }
    ci.write_record();			
    if(ci.do_plugin && ci.total_bytes>0) plugin_process(ci.tempfile_path);

    /* Output the sector hashes to text or XML file if requested */
    if(opt_compute_sector_hashes){
	int count = 1;
	for(vector<string>::const_iterator it = ci.sectorhashes.begin(); it!=ci.sectorhashes.end(); it++){
	    if(sector_bloom){
		sector_bloom->addHex(*it);
		printf("adding %s\n",(*it).c_str());
	    }
	    if(opt_print_sector_hashes){
		if(t) fprintf(t,"sectorhash: %s %d\n",(*it).c_str(),count);
		if(x) x->xmlout("sectorhash",*it,"",false);
	    }
	    count++;
	}
    }

    /* END of file processing */
    if(x) x->pop();
    if(t) fputs("\n",t);
    return TSK_WALK_CONT;
}


/**
 * file name walk callback.  Walk the contents of each file 
 * that is found.
 */
static TSK_WALK_RET_ENUM
dir_act(TSK_FS_FILE * fs_file, const char *path, void *ptr)
{
    /* Ignore NTFS System files */
    if ((TSK_FS_TYPE_ISNTFS(fs_file->fs_info->ftype))
        && (fs_file->name->name[0] == '$'))
        return TSK_WALK_CONT;

    /* If the name has corresponding metadata, then walk it */
    if (fs_file->meta) {
	process_tsk_file(fs_file, path);
    }

    return TSK_WALK_CONT;
}


int proc_fs(TSK_IMG_INFO * img_info, TSK_OFF_T start)
{
    TSK_FS_INFO *fs_info;

    comment("fs start: %" PRIuDADDR, start);
    if(x){
	char buf[1024];
	snprintf(buf,sizeof(buf),"offset='%"PRIuDADDR"'",start);
	x->push("volume",buf);
    }

    /* Try it as a file system */
    if ((fs_info = tsk_fs_open_img(img_info, start, TSK_FS_TYPE_DETECT)) == NULL) {
	partition_info("TSK_Error",tsk_error_get());
        /* We could do some carving on the volume data at this point */
	if(x) x->pop();
        return 1;
    }

    current_partition_num++;
    current_partition_start = fs_info->offset;
    partition_info("Partition_Offset",fs_info->offset);
    partition_info("block_size",fs_info->block_size); sector_size = fs_info->block_size;

    partition_info("ftype",fs_info->ftype);
    partition_info("ftype_str",tsk_fs_type_toname(fs_info->ftype));
    partition_info("block_count",fs_info->block_count);
    partition_info("first_block",fs_info->first_block);
    partition_info("last_block",fs_info->last_block);
    if(t) fputc('\n',t);

    int ret = 0;
    /* Walk the files, starting at the root directory */
    int dir_walk_flags = 
      TSK_FS_DIR_WALK_FLAG_ALLOC |   
      TSK_FS_DIR_WALK_FLAG_UNALLOC | 
      TSK_FS_DIR_WALK_FLAG_RECURSE;

    if (opt_no_orphans) {
      dir_walk_flags |= TSK_FS_DIR_WALK_FLAG_NOORPHAN;
      dir_walk_flags &= ~TSK_FS_DIR_WALK_FLAG_UNALLOC; // removed unallocated
      partition_info("no_orphans",1);
    }
    if (tsk_fs_dir_walk(fs_info, fs_info->root_inum,
            (TSK_FS_DIR_WALK_FLAG_ENUM) dir_walk_flags, dir_act, NULL)) {
	comment("TSK Error: tsk_fs_dir_walk: ",tsk_error_get());
	ret = 1;
	if(x) x->pop();
        return 1;
    }

    /* We could do some analysis of unallocated blocks at this point...  */

    tsk_fs_close(fs_info);
    if(x) x->pop();
    comment("end of volume");
    return ret;
}

/**
 * Volume system walk callback function that will analyze 
 * each volume to find a file system.
 */
static TSK_WALK_RET_ENUM
vs_act(TSK_VS_INFO * vs_info, const TSK_VS_PART_INFO * vs_part, void *ptr)
{
    if (proc_fs(vs_info->img_info, vs_part->start * vs_info->block_size)) {
        // if we return ERROR here, then the walk will stop.  But, the 
        // error could just be because we looked into an unallocated volume.
        // do any special error handling / reporting here.
        tsk_error_reset();
        return TSK_WALK_CONT;
    }

    return TSK_WALK_CONT;
}


/* do_dimage:
 * Process the partition tables in the disk image pointed to by img.
 * File system analysis will be performed on each partition.
 *   mm_open() - opens partition
 *   mm_part_walk() - walks partition
 * If we can't open as a disk image, try to open as a raw file system
 *
 * returns 1 on error and 0 on success
 */
int proc_vs(TSK_IMG_INFO * img_info)
{
    TSK_VS_INFO *vs_info;
    int start = 0;

    // USE mm_walk to get the volumes 
    if ((vs_info = tsk_vs_open(img_info, start, TSK_VS_TYPE_DETECT)) == NULL) {

        /* There was no volume system, but there could be a file system.
	 * Look for one at well-known locations
	 */
        tsk_error_reset();
        if (proc_fs(img_info, 0)==0) return 0;
        tsk_error_reset();
	if (proc_fs(img_info, 63*512)==0) return 0;
        tsk_error_reset();

	/* Just try them all */
	for(int i=1;i<63;i++){
	    if(proc_fs(img_info,i*512)==0) return 0;
	    tsk_error_reset();
	}
	/* Give up */
	return 1;
    }
    else {
        if (tsk_verbose) fprintf(stderr, "Volume system open, examining each\n");

        /* Walk the allocated volumes (skip metadata and unallocated volumes) */
        if (tsk_vs_part_walk(vs_info, 0, vs_info->part_count-1,
                (TSK_VS_PART_FLAG_ENUM) (TSK_VS_PART_FLAG_ALLOC), vs_act, NULL)) {
            tsk_vs_close(vs_info);
            return 1;
        }
        tsk_vs_close(vs_info);
    }
    return 0;
}



void process_bloomfile(const char *fname)
{
    if(fname[0]!='+' && fname[0]!='-'){
	fprintf(stderr,"usage: -b[+/-]bloomfile.bloom\n");
	usage();
    }
    bloomset_element::Action action = (fname[0]=='+' ? bloomset_element::INCLUDE : bloomset_element::EXCLUDE);
    bloomset_element *be = new bloomset_element(action, fname+1);
    mybloomset.push_back(be);
}


void process_scalpel_audit_file(TSK_IMG_INFO *img_info,const char *audit_file)
{
    const char *fmt = "%12s%qd\t\t%3s%zd";
    FILE *f = fopen(audit_file,"r");
    while(!feof(f)){
	char buf[1024];
	while(fgets(buf,sizeof(buf),f)){
	    char filename[13];
	    int64_t start;
	    char chop[4];
	    size_t length;
	    memset(filename,0,sizeof(filename));
	    memset(chop,0,sizeof(chop));
	    if(sscanf(buf,fmt,filename,&start,chop,&length)==4){
		/* See if we can get the bytes */

		if(a) a->new_row();
		if(x) x->push("fileobject"); 

		content ci;
		ci.evidence_dirname = "?/";
		ci.set_filename(filename); // fictitious filename, but needed for plugins

		char *buf2 = (char *)calloc(length,1);
		size_t r2 = tsk_img_read(img_info,(TSK_OFF_T)start,buf2,length);

		file_info("filesize",r2);
		if(r2!=length){
		    file_info("carvelength",length);
		}

		ci.add_seg(start,start,0,r2,TSK_FS_BLOCK_FLAG_RAW);	// may not be able to read it all
		ci.add_bytes(buf2,0,r2); 
		ci.write_record();
		free(buf2);

		/* END of file processing */

		if(x) x->pop();
		if(t) fputs("\n",t);
	    }
	}
    }
}

bool has_unprintable(const u_char *buf,int buflen)
{
    while(buflen>0){
	if(!isprint(*buf)) return true;
	buf++;
	buflen--;
    }
    return false;
}

void sig_info(int arg)
{
    if(a){
	printf("a=%p\n",a);
	printf("\n");
    }
}


#ifdef HAVE_AF_DISPLAY_AS_QUAD
static const char *quads[] = {
    AF_IMAGESIZE,
    AF_BADSECTORS,
    AF_BLANKSECTORS,
    AF_DEVICE_SECTORS,
    0
};


int af_display_as_quad(const char *segname)
{
    for(int i=0;quads[i];i++){
	if(strcmp(segname,quads[i])==0) return true;
    }
    return false;
}


int af_display_as_hex(const char *segname)
{
    if(strcmp(segname,AF_MD5)==0) return 1;
    if(strcmp(segname,AF_SHA1)==0) return 1;
    if(strcmp(segname,AF_SHA256)==0) return 1;
    if(strcmp(segname,AF_IMAGE_GID)==0) return 1;
    return 0;
}
#endif


int main(int argc, char * const *argv)
{
    int ch;
    extern int optind;
    const char *arff_fn = 0;
    const char *text_fn = 0;
    const char *xml_fn = 0;
    const char *audit_file = 0;
    bool opt_x = false;
    string command_line;

    int t0 = time(0);

    /* Make a copy of the command line */
    for(int i=0;i<argc;i++){
	if(i>0) command_line.push_back(' ');
	command_line.append(argv[i]);
    }

    while ((ch = getopt(argc, argv, "a:mMmdv1s:A:EX:T:n:c:b:2B:S:VfFC:xgOzh")) > 0) {
	switch (ch) {
	case '1': opt_sha1++;break;
	case 'm':
	    opt_body_file = 1;
	    opt_sha1 = 0;
	    opt_md5  = 1;
	    t = stdout;
	    break;
	case 'A': arff_fn = optarg;break;

	    /* Sector hash stuff */
	case 'B': 
	    opt_compute_sector_hashes=true;
	    sector_bloom = new NSRLBloom();
	    if(sector_bloom->create(optarg,160,30,4,"Sector hash")){
		err(1,"%s",optarg);
	    }
	    break;
	case 'E':
	    opt_print_sector_hashes = true;
	    opt_compute_sector_hashes=true;
	    break;
	case 'S': sectorhash_size = atoi(optarg); break;

	    /***/
	case 'C': file_count_max = atoi(optarg);break;
	case 'T': text_fn = optarg;break;
	case 'O': opt_no_orphans=true; break;
	case 'V': print_version();exit(0);
	case 'x': opt_x = true;break;
	case 'X': xml_fn = optarg;break;
	case 'a': audit_file = optarg;break;
	case 'b': process_bloomfile(optarg);break;
	case 'c': config_file = optarg; break;
	case 'd': opt_debug++; break;
	case 'g': opt_get_fragments++; break;
	case 'f': opt_magic = true;break;
	case 'F': opt_magic = false;break;
	case 'M': opt_md5++; break;
	case 'n': namelist.push_back(optarg);break;
	case 's': save_outdir = optarg; opt_save = true; break;
	case 'v': tsk_verbose++; break; 			// sleuthkit option
	case 'z': opt_sha1=false;opt_md5=false;break;
	    
	case '2': opt_multithreaded = false; break;

	case 'h':
	case '?':
	default:
	    fprintf(stderr, "Invalid argument: %s\n", argv[optind]);
	    usage();
	}
    }

    if (optind >= argc) usage();
    argc -= optind;
    argv += optind;

    const char *filename = argv[0];
    if(!filename){
	errx(1,"must provide filename");
    }

    if(opt_save){
	if(access(save_outdir.c_str(),F_OK)){
	    if(mkdir(save_outdir.c_str(),0777)){
		err(1,"Cannot make directory: %s",save_outdir.c_str());
	    }
	}
	if(access(save_outdir.c_str(),R_OK)){
	    err(1,"Cannot access directory: %s",save_outdir.c_str());
	}
    }

    if(text_fn){
	if(access(text_fn,F_OK)==0) errx(1,"%s: file exists",text_fn);
	t = fopen(text_fn,"w");
	if(!t) err(1,"%s",text_fn);
    }

    if(arff_fn){
	if(access(arff_fn,F_OK)==0) errx(1,"%s: file exists",arff_fn);
	a = new arff("fiwalk");		// the ARFF output object
	a->set_outfile(arff_fn);
    }

    /* XML initialization */

    if(opt_x){
	x = new xml();
	x->set_outFILE(stdout);
    }
    if(xml_fn){
	if(x) errx(1,"Cannot write XML to stdout and file at same time\n");
	if(access(xml_fn,F_OK)==0) errx(1,"%s: file exists",xml_fn);
	x = new xml();
	x->set_outfilename(xml_fn);
	x->set_makeDTD(true);
    }
	
    /* If no output file has been specified, output text to stdout */
    if(a==0 && x==0 && t==0){
	t = stdout;
    }

    if(config_file){
	comment("Reading configuration file %s",config_file);
	config_read(config_file);    /* Read the configuration file */
    }

    if(strstr(filename,".aff") || strstr(filename,".afd") || strstr(filename,".afm")){
#ifndef HAVE_LIBAFFLIB
	fprintf(stderr,"ERROR: fiwalk was compiled without AFF support.\n");
	exit(0);
#else
	if((tsk_img_type_supported() & TSK_IMG_TYPE_AFF_AFF)==0){
	    fprintf(stderr,"ERROR: fiwalk was compiled with AFF support but the TSK library is not.\n");
	    exit(0);
	}
#endif
    }

    /* If we are outputing ARFF, create the ARFF object and set the file types for the file system metadata */
    if(a){
	a->add_attribute("id",arff::NUMERIC);
	a->add_attribute("partition",arff::NUMERIC);
	a->add_attribute("filesize",arff::NUMERIC);
	a->add_attribute("mtime",arff::DATE);
	a->add_attribute("ctime",arff::DATE);
	a->add_attribute("atime",arff::DATE);
	a->add_attribute("fragments",arff::NUMERIC);
	a->add_attribute("frag1startsector",arff::NUMERIC);
	a->add_attribute("frag2startsector",arff::NUMERIC);
	a->add_attribute("filename",arff::STRING);
	if(opt_md5) a->add_attribute("md5",arff::STRING);
	if(opt_sha1) a->add_attribute("sha1",arff::STRING);
    }

    /* output per-run metadata for XML output */
    if(x){
	/* Output Dublin Core information */
	x->open();
	x->push("fiwalk","xmloutputversion='0.3'");
	x->push("metadata",
		"\n  xmlns='http://afflib.org/fiwalk/' "
		"\n  xmlns:xsi='http://www.w3.org/2001/XMLSchema-instance' "
		"\n  xmlns:dc='http://purl.org/dc/elements/1.1/'" );
	x->xmlout("dc:type","Disk Image",empty,false);
	x->pop();
	    
	/* Output carver information per photorec standard */
	x->push("creator");
	x->xmlout("program","fiwalk");
	x->xmlout("version",PACKAGE_VERSION);
	x->push("build_environment");
	x->xmlprintf("compiler","","GCC %d.%d",__GNUC__, __GNUC_MINOR__);
	x->xmlout("compilation_date",__DATE__ " " __TIME__);
	x->xmlout("library", "", string("name=\"tsk\" version=\"")
		  + tsk_version_get_str() + "\"",false);
	x->xmlout("os",UNAMES,"",false);
#ifdef HAVE_LIBAFFLIB
	x->xmlout("library", "", string("name=\"afflib\" version=\"")
		  + af_version() +"\"",false);
#endif
	x->pop();
	x->push("execution_environment");
#ifdef HAVE_SYS_UTSNAME_H
	struct utsname name;
	if(uname(&name)==0){
	    x->xmlout("os_sysname",name.sysname);
	    x->xmlout("os_release",name.release);
	    x->xmlout("os_version",name.version);
	    x->xmlout("host",name.nodename);
	    x->xmlout("arch",name.machine);
	}
#endif	
	x->xmlout("command_line", command_line); // quote it!
	x->xmlprintf("uid","","%d",getuid());
	x->xmlout("username",getpwuid(getuid())->pw_name);
	x->pop();			// <execution_environment>
	x->pop();			// creator
    }

    /* Check that we have a valid file format */
    if(x) x->push("source");
    partition_info("imagefile",filename);

    if(!x){
	partition_info("fiwalk_version",PACKAGE_VERSION);
	partition_info("start_time",mytime());
	partition_info("tsk_version",tsk_version_get_str());
    }
    if(x) x->pop();

    if (opt_debug) printf("calling tsk_img_open(%s)\n",filename);

#ifdef HAVE_LIBAFFLIB
    /* Get all of the AFF metadata and put into the XML file... */
    AFFILE *af = af_open(filename,0,O_RDONLY);
    if(!af) af_err(1,"af_open(%s)",filename);
    
    if(!x) partition_info("aff_version",af_version());

    aff::seglist sl(af);

    for(aff::seglist::const_iterator seg = sl.begin();seg != sl.end(); seg++){
	unsigned long arg;
	size_t datalen = 0;
	char hashname[256];			// 
	
	const char *segname = seg->name.c_str();

	if(segname[0]==0) continue;	// skip this segment
	if(af_segname_page_number(segname)!=-1) continue; // don't get the data pages
	if(af_segname_hash_page_number(segname,hashname,sizeof(hashname))!=-1) continue; // don't do hash pages
	if(af_is_signature_segment(segname)) continue; // don't do signature segments
	if(datalen>65536) continue;	// don't include segments larger than this
	/* Now get the data */
	u_char *data = (u_char *)malloc(datalen+1);
	if(af_get_seg(af,segname,&arg,data,&datalen)){
	    free(data);
	    continue;			// shouldn't have had an error getting it...?
	}
	data[datalen] = 0;
	/* Now figure out how to code the XML */
	if(datalen==0){
	    partition_info(segname,arg);
	} else if(af_display_as_quad(segname)){
	    char display[20];
	    int64_t quad;
	    af_get_segq(af,segname,&quad);
	    snprintf(display,sizeof(display),"%"PRId64,quad);
	    partition_info(segname,quad);
	} else if(af_display_as_hex(segname)){
	    int bhexsize = datalen*2+2;
	    char *bhexbuf = (char *)calloc(bhexsize,1);
	    partition_info(segname,af_hexbuf(bhexbuf,bhexsize,data,datalen,AF_HEXBUF_NO_SPACES),"coding='base16'");
	    free(bhexbuf);
	} else if(has_unprintable(data,datalen)){
	    int  b64size = datalen*2+2;
	    char *b64buf = (char *)calloc(b64size,1);
	    b64_ntop(data,datalen,b64buf,b64size);
	    partition_info(segname,b64buf,"coding='base64'");
	    free(b64buf);
	} else{
	    partition_info(string((const char *)segname),string((const char *)data));
	}
	free(data);
    }
    sector_size = af_get_sectorsize(af);
    af_close(af);
#endif    

#ifdef SIGINFO
    signal(SIGINFO,sig_info);
#endif
    TSK_IMG_INFO *img_info;
    img_info = tsk_img_open_utf8(argc,(const char **)argv,TSK_IMG_TYPE_DETECT);
    if (!img_info){
	comment("TSK Error (img_open) %s",tsk_error_get());
	comment("=EOF= (img_open)");
	exit(1);
    }

    if(audit_file){
	comment("audit file: %s",audit_file);
	process_scalpel_audit_file(img_info,audit_file);
    }
    else{
	if (opt_debug) printf("calling do_dimage()\n");
	if (proc_vs(img_info)){
	    comment("TSK Error (do_dimage) %s",tsk_error_get());
	}
    }
    tsk_img_close(img_info);

#ifdef HAVE_GETRUSAGE
    /* Print usage information */
    struct rusage ru;
    memset(&ru,0,sizeof(ru));
    int t1 = time(0);
    comment("clock: %d",t1-t0);
    if(getrusage(RUSAGE_SELF,&ru)==0){
	if(x) x->push("runstats");
	partition_info("user_seconds",ru.ru_utime.tv_sec);
	partition_info("system_seconds",ru.ru_stime.tv_sec);
	partition_info("maxrss",ru.ru_maxrss);
	partition_info("reclaims",ru.ru_minflt);
	partition_info("faults",ru.ru_majflt);
	partition_info("swaps",ru.ru_nswap);
	partition_info("inputs",ru.ru_inblock);
	partition_info("outputs",ru.ru_oublock);
	partition_info("stop_time",cstr(mytime()));
	if(x) x->pop();
    }
#endif

    // *** Added <finished time="(time_t)" duration="<seconds>" />

    if(a){
	a->write();
	delete a;
    }
    if(t) comment("=EOF=");
    if(x) {
	x->pop();
	x->close();
	delete x;
    }
    exit(0);
}
